/*************************************************************
wasm_vensim.h Copyright Ventana Systems, Inc.
*********************************************/

#ifndef _WASM_FUNCTIONS_H_
#define _WASM_FUNCTIONS_H_

#if defined(__cplusplus)
#define C_LINKED_FUNCTION extern "C"
#else //#if defined(__cplusplus)
#define C_LINKED_FUNCTION 
#endif //#if defined(__cplusplus)


long Get_NUM_VARS();
long Get_NUM_LEVELS();
long Get_NUM_DELAYS();
long Get_NUM_AUX();
long Get_NUM_LOOKUPS();
long Get_NUM_STRINGS();
charutf8 *Get_VarName(long nIndex);
charutf8 *Get_LookupName(long nIndex);
int Get_VarType(long nIndex);
void InitConstants();
void SetupExternalData();
void comp_rate(void);
void comp_delay(void);
void init_time(void);
void init_tstep(void);
void comp_init(void);
void comp_reinit(void);
void comp_tstep(void);
void comp_aux(void);

void SetDebugModeOff();
void SetDebugModeOn();


C_LINKED_FUNCTION void PrintDebugMessage(char *fmt, ...);

#define MAX_DEBUG_MESSAGE_LEN 512

//WASM exported prototypes
charutf8 *Get_VarName(long nIndex);
charutf8 *Get_LookupName(long nIndex);
int Get_VarType(long nIndex);
int GetConstantUsingIndex(long nIndex, double *pResult);
int GetConstantUsingVarName(charutf8 *strVarName, double *pResult);
double GetConstant(charutf8 *strVarName);
int SetConstant(long nIndex, charutf8 *strVarName, double fValue);
int SetStringVar(int nIndex, charutf8 *strVarName, charutf8 *strValue);
int SetLookup(long nIndex, charutf8 *strVarName, long nNumValues, REAL *fxValues, REAL *fyValues);
int SetValue(long nIndex, charutf8 *strVarName, double fValue, int nValueType);
int SetGamingVar(long nIndex, charutf8 *strVarName, double fValue);
C_LINKED_FUNCTION int SetDataParameters(long nvar, long nDataCount, charutf8 *strModelVarName, charutf8 *strVarNameToRetrieve, int nFileType, charutf8 *filename, charutf8 *strSheetName, charutf8 *strTimeRowOrCol, charutf8 *strStartPosition, charutf8 *strSlope, int nFromVDFFlag, int ninterpolation_pattern_type, int nFunctionIndex, int nLookupFlag);
void init_exog_data_var_manager();
void comp_exog_data_var_manager();
C_LINKED_FUNCTION int setup_exog(int *efileid, charutf8 *efile_source, int num_file, int maxtimes, int forpayoff);

static void vec_arglist_init();

VECTOR_ARG* vecarg_assign(int ind, COMPREAL *vals);

void FreeDIMs();
void SetupDIMS();
void CreateDIMS(long nNumDims);
void SetDIM_INFO(int nIndex, int nFam0, int nFam1, int nFam2, int nFam3, int nFam4, int nFam5, int nFam6, int nFam7, int dim0, int dim1, int dim2, int dim3, int dim4, int dim5, int dim6, int dim7, int dim_vol0, int dim_vol1, int dim_vol2, int dim_vol3, int dim_vol4, int dim_vol5, int dim_vol6, int dim_vol7, int tot_dim, int tot_vol);

int VEFCC add_vecarg_info(ST_INDEX *stlist, ST_INDEX *pDIMOffsets, int n);

#endif

