/*************************************************************
  wasm_vensim.h Copyright Ventana Systems, Inc.
 *********************************************/

#ifndef _WASM_VENSIM_STRUCTURES_H_
#define _WASM_VENSIM_STRUCTURES_H_

#ifdef __cplusplus
extern "C"
{
#endif //#ifdef __cplusplus

 /* combine the MARKET_MATCH vectors into a struct to simplify
 passing it and also to make it easier to allocat and free the
 temporary storage needed to run the algorithm - note that
 the elements of this are all stored as vectors even though most
 are 3 dimensional matrices - use the BSC etc macros to compute
 offsets for indices */
typedef struct {
	int max_allocp_size; /* control of allocp */
	int nbuyer; /* maxes for allocation */
	int nseller;
	int nchannel;
	int nsellerp; /* number of priority entries in sp_supplyp */
	int nbuyerp; /* number of priority entries in the sbcp_priority matrix */
	int nbuyer_attractp; /* number of priority entries in bscp_attract */
	int nseller_priorityp; /* number of priority entries in sbcp_priority */

						   /* these are transferred by address from parameters sent or temporarily
						   allocated and filled in to simplify calling protocol */
	COMPREAL *bsc_sales; /* by buyer, seller, channel */
	COMPREAL *bsc_adjuster; /* allows demand from different buyers to mean something different to sellers */
	COMPREAL *bsc_aware;
	COMPREAL *bscp_attract;/* attractiveness of suppliers by buuyers - last dimension is priority profile  */
	COMPREAL *sbcp_priority;
	COMPREAL *b_demand;
	COMPREAL *bp_demandp;
	COMPREAL *s_supply;
	COMPREAL *sp_supplyp;

	/* below are allocated */
	int *order; /* used in the old ALLOC_P function only */

	COMPREAL *sc_temp_allocation;
	COMPREAL *b_demand_shortfall;
	COMPREAL *sc_temp_sales;
	COMPREAL *bc_temp_allocation;
	COMPREAL *bc_temp_sales;
	COMPREAL *bsc_old_sales; /* for comparison with current comps */
	COMPREAL *s_excess_supply; /* for ongoing rounds - keep the offer open */
} MARKET_MATCH_PARMS;


typedef struct tag_store_pairstruct
{
	REAL time; // time to emit next value
	REAL val;  // value at emmision
} STORE_PAIRSTRUCT;

typedef struct tag_comp_pairstruct
{
	COMPREAL time; // time to emit next value 
	COMPREAL val;  // value at emmision 
} COMP_PAIRSTRUCT;


#define COMMON_DELAY_STRUCT_VARTYPE		V8BYTES
#define MAX_LEVELS_STORED				20000000
#define MAX_LEVELS_STORED_WARNING_THRESHOLD	5000000

typedef struct tag_common_delay_struct
{
	VAL_OFF off; /* from GV.LEVEL for value */
	V2BYTES type; /*uses the function type from sim.h except 0 is a synonym for F_DELAY_FIXED 1 for F_DELAY_MATERIAL and 2 for F_DELAY_INFORMATION */
	union
	{
		COMMON_DELAY_STRUCT_VARTYPE num_pairs; /* the number of pairs active */
		COMMON_DELAY_STRUCT_VARTYPE order; /* for DELAY_N and SMOOTH N */
	} u3;

	COMMON_DELAY_STRUCT_VARTYPE		max_pairs; /* the maximum #pairs without reallocation */
	union
	{
		COMMON_DELAY_STRUCT_VARTYPE live_passed; /* flag to indicate that initialization is passed as int for alignment reasons */
		COMMON_DELAY_STRUCT_VARTYPE last_was_transition; /* for depreciation function last was a transition */
	} u4;
} COMMON_DELAY_STRUCT;



typedef struct tag_store_delay_struct
{
	union
	{
		REAL delay_time; /* for fixed length delay */
		REAL profile_volume; /* for delay_profile */
		REAL granularity; /* for internal ROR */
		REAL fiscal_period;/* for NPV function */
	} u1;
	union
	{
		REAL init_val; /* the value to use before first live value available */
					   /* for some this is used to initialize level sequences on the first active compuation */
		REAL position; /* for conveyors - stuff after position-time_step/delay_time drops off new stuff is put on at position-1 after position is updated - all inputs are paired (one wasted number) */
		REAL base_time; /* for internal ROR the lowest time used anywhere */
		REAL fiscal_offset;   /* for depreciation schedules */
	} u1a;
	union
	{
		REAL missing_val; /* for material - the value to use when no live values are available, but at least one live value has been seen for information the last value returned */
		REAL outval;  /* for DELAY_N, SMOOTH_N - next emergent val */
		REAL time_mult; /* used to stretch the profile so that the mean in the delay time */
		REAL lastinflow; /* for memory functions - need to back this out */
	} u2;
	COMMON_DELAY_STRUCT c;
	STORE_PAIRSTRUCT pairs[1];
} STORE_DELAY_STRUCT;


typedef struct tag_comp_delay_struct
{
	union
	{
		COMPREAL delay_time; /* for fixed length delay */
		COMPREAL profile_volume; /* for delay_profile */
		COMPREAL granularity; /* for internal ROR */
		COMPREAL fiscal_period;/* for NPV function */
	} u1;
	union
	{
		COMPREAL init_val; /* the value to use before first live value available */
						   /* for some this is used to initialize level sequences on the first active compuation */
		COMPREAL position; /* for conveyors - stuff after position-time_step/delay_time drops off new stuff is put
						   on at position-1 after position is updated - all inputs are paired (one wasted number) */
		COMPREAL base_time; /* for internal ROR the lowest time used anywhere */
		COMPREAL fiscal_offset;   /* for depreciation schedules */
	} u1a;
	union
	{
		COMPREAL missing_val; /* for material - the value to use when no live values
							  are available, but at least one live value has been seen
							  for information the last value returned */
		COMPREAL outval;  /* for DELAY_N, SMOOTH_N - next emergent val */
		COMPREAL time_mult; /* used to stretch the profile so that the mean in the delay time */
		COMPREAL lastinflow; /* for memory functions - need to back this out */
	} u2;
	COMMON_DELAY_STRUCT c;
	COMP_PAIRSTRUCT pairs[1];
} COMP_DELAY_STRUCT;

typedef struct tag_dim_info_vdfx
{
	BLK_OFFSET next;

	union
	{
		V4BYTEU tot_vol; /* the total size  */
		BLK_OFFSET parent; /* the parent dimensions for exceptions */
	} u1;

	union
	{
		struct
		{
			//V4BYTEU PADDING;
			VAL_OFF dim[MAX_DIM];     /* total size of a dimension */
			VAL_OFF dim_vol[MAX_DIM]; /* the total volume for one subscript */
		} d;
		struct
		{
			char *except_flag_ptr; /* for large one - scr_alloced so not permanent */
			VAL_OFF except_off[2 * MAX_DIM - 1]; /* for small number of exceptions */
		} e;
	} u2;
	ST_INDEX fam[MAX_DIM];   /* families for elements - actual entries for excemptions : TONY. I think this is a symbol table entry */
	V2BYTES tot_dim; /* number of elements */
	V2BYTES except_type; /* the number of exceptions, but capped at 30000 */
} DIM_INFO_VDFX;

#define DIM_INFO DIM_INFO_VDFX


typedef struct
{
	union
	{
		COMPREAL *val;
		ST_INDEX stindx; /* used by CRTC.c to provide info for a call to assign_vecarg */
	} u; /* not unioned in external stuff */

	COMPREAL *firstval;
	DIM_INFO *dim_info;
	charutf8 *varname;
} VECTOR_ARG;

typedef struct {
	COMPREAL *xvals;
	COMPREAL *yvals;
	long nvals;
} DATA_VECTOR;

typedef struct {
	V4BYTEU keyval;
	charutf8 *varname;
	COMPREAL *timevals;
	union {
		COMPREAL **vals;
		DATA_VECTOR **dv;
	} u;
	long nvar;
	long ntime;
} DATA_MATRIX;

typedef struct
{
	unsigned long keyval;
	char *varname;
	COMPREAL **vals;
	long ncol;
	long nrow;
} CONSTANT_MATRIX;

#ifdef WASM
#define of2tabval(off) off
#define CFUNCTION
#endif	//#ifdef WASM


#ifdef VDFX
typedef	float				LOOKUP_TABLE_VAL_TYPE;	//Tony. This was an attempt to use double as the standard for lookup tables.
													//The tab values seem to be read in ok, but once they are used, things go wrong (search ARRAY_CONTAINING_TAB_VALUES for where it's stored, 
													//and GV.TAB[offi].base = of2tabval( GV.TAB[offi].y )[0]; in sim.c for where it goes wrong when this is changed to a double.
#else
typedef	float				LOOKUP_TABLE_VAL_TYPE;
#endif


typedef struct tag_table_range_VDFX
{
#ifdef USE_OFFSET_HANDLES
	VAL_OFF history;
#else
#ifndef BUILD_x64
	V4BYTES x86_Padding_01;
#endif
	charutf8 *history;
#endif
	VAL_OFF first;
	VAL_OFF tot_vol;
	BLK_OFFSET next;
} TABLE_RANGE_VDFX;

#define TABLE_RANGE TABLE_RANGE_VDFX

// note that this is really much too big we only need it for simulation but it is actually written to disk 
typedef struct tag_tab_type_VDFX
{
	LOOKUP_TABLE_VAL_TYPE base; /* the base val */
	LOOKUP_TABLE_VAL_TYPE xgap; /* the time gap to the next x value */
	LOOKUP_TABLE_VAL_TYPE ygap; /* the value gap the the next y value */
	LOOKUP_TABLE_VAL_TYPE lstx; /* the x value of the  base */
	LOOKUP_TABLE_VAL_TYPE nextx; /* the x value of the next entry after base */
#ifdef WASM
	LOOKUP_TABLE_VAL_TYPE *x;// pxVals; /* pointer to the x values */
	LOOKUP_TABLE_VAL_TYPE *y;// pyVals; /* pointer to the y values */
#else
	BLK_OFFSET x; /* pointer to the x values */
	BLK_OFFSET y; /* pointer to the y values */
#endif
	V4BYTES curind; /* the current index value (matches base) */
	V4BYTES nNumValsAndTableType; /* the number of value for the tabe */
	TABLE_RANGE range[1]; /* used for reporting over under and in messages */
} TAB_TYPE_VDFX;

#define TAB_TYPE TAB_TYPE_VDFX


typedef union
{
	COMPREAL val;
	VECTOR_ARG *vec;
	TAB_TYPE *tab;
	const unsigned char *literal;
	CONSTANT_MATRIX *constmat;
	DATA_MATRIX *datamat;
} VV;

/*
typedef struct tag_st_head_struct
{
	VAL_OFF num_lev;
	VAL_OFF num_delay;
} ST_HEAD;
*/

#define MAX_SUB_NAME_LEN 512 /* the maximum lenght of a fully subscripted variable */

#ifdef WASM_INCLUDE_DATA
#include "d:/Vensim/Vensim/src/vensim/exog_data_structs.h"
#endif

#define NUM_ERROR_TYPES 14	//This is VENSIM_ERROR_LAST_ERROR_INDEX in ventype.h

typedef struct tag_WASM_glob_vars
{
	VAL_OFF lastpos ;
	COMPREAL *LEVEL ;
	COMPREAL *RATE ;
	TAB_TYPE *TAB ;
//	const ST_HEAD *sinfo ;
	COMPREAL time ;
	COMPREAL time_step ;
	COMPREAL time_plus ;
	COMPREAL time_minus ;
	COMPREAL final_time ;
	COMPREAL initial_time ;
	COMPREAL save_time ;
	COMPREAL saveper ;
	VECTOR_ARG *vector_arglist;
	long nNumVectorArgs;	//Number of vectors in *vector_arglist
	char simulation_state ;
	const char gaming_state ;
	const char resume_flag;
	const char integ_type ;
	const char simulation_type ;
	const int fs_override_active ;
	const COMPREAL *fs_override_vals ;
	const char *fs_overridden ;
	COMP_DELAY_STRUCT **delay;
	void *ran_vals_blk;
	int ran_num_gen;
	long seed;
	DIM_INFO	*pDIMs;

	long g_NUM_VARS;
	long g_NUM_LOOKUPS;
	long g_NUM_LEVELS;
	long g_NUM_DELAYS;
	long g_NUM_STRINGS;
	long g_NUM_AUX;

	MARKET_MATCH_PARMS *mm_parms; // used in ALLOCP, ALLOCATION... MARKET... 

	COMPREAL VLARGE *lev_extra[MAX_LEV_EXTRA];
	int max_broydn_n; /* used in zero solving */

	charutf8 *curpos;
	charutf8 *wide_inline; // for wide lines using continuations 
	charutf8 token[MAX_SUB_NAME_LEN];
	charutf8 readline[MAX_LINE_LEN + 1]; // for reading things in

	COMPREAL game_play_stop;
	COMPREAL game_play_interval;

	unsigned short *up_to_low;
	unsigned short *low_to_up;
	charutf8 *char_char;
	unsigned short *char_order;

	int ntimes;
	int maxtimes;
	COMPREAL *vtime;

	//Exog data vars
	#ifdef WASM_INCLUDE_DATA
	COMPREAL data_granularity;
	int exg_cind;
	EXG_INF *fexg_elm; /* the first exogenous element */
	EXG_INF *curexg_elm; /* the first exogenous element */
	VAL_OFF n_exg_elm; // number of data elements
	char data_want_recompute; // recompute the data to respond to constant changes
	char data_want_cache; // use cache - may get invalidated by time axis changes
	int data_n_runs; // number of runs on current cache reset
	DATA_CACHE *data_cache;
//	EXG_ODBC_DATA_SRC *exg_src;

	int nbasetimes;
	COMPREAL *vbasetime; /* used to extend the vtime base with time shifted values */
	COMPREAL *vtemptime; /* just a databuffer */

	COMPREAL time_slop; /* allowable distance before two time points are merged */
	#endif //WASM_INCLUDE_DATA

	int error_count[NUM_ERROR_TYPES]; /* number of errors of different types made */
	//void *pRunManager;
	void *pStringVarManager;

} WASM_GLOB_VARS ;

#if !defined(SFMT_MEXP)
#if defined(__GNUC__) && !defined(__ICC)
//#warning "SFMT_MEXP is not defined. I assume MEXP is 19937."
#define SFMT_MEXP 19937
#endif
#define SFMT_MEXP 19937
#endif

#ifndef SFMTST_H
#define SFMT_N (SFMT_MEXP / 128 + 1)
union W128_T
{
	uint32_t u[4];
	uint64_t u64[2];
};

/** 128-bit data type */
typedef union W128_T w128_t; 
struct SFMT_T
{
	/** the 128-bit internal state array */
	w128_t state[SFMT_N];
	/** index counter to the 32-bit internal state array */
	int idx;
};

typedef struct SFMT_T sfmt_t;
#endif

typedef struct
{
	union
	{
		struct
		{
			float r[98];
			int32_t ix1, ix2, ix3;
		} norm;
		struct
		{
			int32_t ma[56];
			int inext, inextp;
		} sub;
		struct
		{ // mersenne twister state
			sfmt_t state;
		} mt;
	} rantype;

	float gset;
	char random_normal_set;
} RANSTATE;


typedef struct tag_random_control
{
	double(*random_function)(struct tag_random_control *);
	BigREAL streamid;			// uses == to validate the id and (long) to get a seed 
	int32_t seed;				// storage for seed used; normally seed = streamid, but could differ for randomized values
	RANSTATE *oldstate;		// this is to hold for RK integration to prevent changes
	RANSTATE ranstate;
	char fromopt;
	struct tag_random_control *next;
} RANDOM_CONTROL;

typedef struct minvert_tag {
	COMPREAL *tempmat;
	COMPREAL *scratch;
	COMPREAL *col;
	int *indx;
} MINVERT;

typedef struct brzero_tag {
	COMPREAL tolerance_z; /* small enough output change not to matter */
	COMPREAL *vi; /* input to vensim code */
	COMPREAL *vibest; /* record whenver there is an improvement */
	COMPREAL *vo; /* output from vensim code */
	COMPREAL *vsi; /* scale factors for input */
	COMPREAL *vso; /* scale factors for output */
	COMPREAL *x; /* scaled input vector - divided by vsi before going to vensim*/
	COMPREAL *xold; /* held value for x */
	COMPREAL *z; /* scaled output vensim out * vso */
	COMPREAL *zold; /* old scaled output vensim out * vso */
	COMPREAL *c; /* for QR decomposition */
	COMPREAL *s; /* working vector for main broydn */
	COMPREAL *t; /* working for broydn */
	COMPREAL *w; /* direction - working for broydn */
	COMPREAL *diagonal; /* for QR decomposition */
	COMPREAL *direction; /* the direction to take steps in */
	COMPREAL *gradient; /* function gradient */
	COMPREAL **jacobian;
	COMPREAL **qt;
	COMPREAL fbest; /* smallest reached so far */
	int vn; /* number of elements */
} BRZERO;

typedef struct WASM__tag_static_vars
{
	RANDOM_CONTROL *First_sim_ran;
	RANDOM_CONTROL *First_opt_ran;
	RANDOM_CONTROL *First_sens_ran;
	MINVERT *Minvert_var;
	BRZERO *Brzero_var;
	int Broken_line;
	int inline_len;
	COMPREAL VLARGE *Mtime_step; /* the time step */
} WASM_STATIC_VARS;

extern WASM_GLOB_VARS	*VGV;
extern WASM_STATIC_VARS	*SS;

#ifdef __cplusplus
}
#endif //#ifdef __cplusplus

#endif


